--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   SimpleTypes.hs
--
-- Author:
--   Jan Hoffmann (2010)
--
-- Description:

module SimpleTypes where


import qualified Data.Map as M
import Data.List
import Control.Monad

import RAMLTypes
import CallGraph (getSCCs)


typeProgram funList e_main = 
   do { let sigma = M.fromList [ (f,t_f) | (f,xs,e_f,t_f) <- funList ]
      ; typedMain <- typeMain e_main sigma
      ; typedFunList <- sequence [ sequence [ typeFun f xs e_f t_f sigma  | (f,xs,e_f,t_f) <- funsSCC ] | funsSCC <- getSCCs funList ]
      ; return (typedMain, typedFunList)
      }



typeMain e sigma = elimNatConst $ typeExp e M.empty sigma
                   


typeFun f xs e_f (FunType tArg tRes) sigma = 
   do { let context = mkContext xs tArg
      ; te_f <- typeExp e_f (M.fromList context) sigma
      ; when (not $ tRes `subTypeOf` (typeOf te_f) )
          ( fail $ "The function " ++ f ++ " is declared to have a result of type " ++ (show tRes)
                   ++ "\nbut the inferred type of the function body is " ++  (show $ typeOf te_f))
      ; return (f,(context,setType te_f tRes))
      }
    where
       mkContext [x] t = [(x,t)]
       mkContext xs  (TTuple ts) = zip xs ts


t1 `subTypeOf` t2 = 
    case (t1,t2) of
      _ | t1 == t2      -> True
      (TInt, TNatConst) -> True
      (TNat, TNatConst) -> True
      (TList _, TNil)   -> True
      (TBTree _, TLeaf) -> True
      (TList t1', TList t2') -> 
          t1' `subTypeOf` t2'
      (TBTree t1', TBTree t2') -> t1' `subTypeOf` t2'
      (TTuple ts, TTuple ts') | length ts == length ts' -> all (uncurry subTypeOf) (zip ts ts')
      _                 -> False


unifyTypes t1 t2 = 
    case (t1,t2) of
      _ | t1 == t2 ->
          return t1
      (TList t, TNil) -> 
          return $ TList t
      (TNil, TList t) -> 
          return $ TList t
      (TNatConst, TInt) -> 
          return $ TInt
      (TInt, TNatConst) -> 
          return $ TInt
      (TNatConst, TNat) -> 
          return $ TNat
      (TNat, TNatConst) -> 
          return $ TNat
      (TBTree t, TLeaf) ->
          return $ TBTree t
      (TLeaf, TBTree t) ->
          return $ TBTree t
      (TList t1', TList t2') -> 
          unifyTypes t1' t2' >>= \t -> return $ TList t
      (TBTree t1', TBTree t2') -> 
          unifyTypes t1' t2' >>= \t -> return $ TBTree t
      (TTuple ts, TTuple ts') | length ts == length ts' -> 
          sequence [unifyTypes t t' | (t,t') <- zip ts ts'] >>= \ts -> return $ TTuple ts
      _ -> fail $ "Cannot unify types " ++ (show t1) ++ " and " ++ (show t2) ++ "."


elimNatConst compTe = 
    do { te <- compTe
       ; case typeOf te of
           TNatConst -> return $ setType te TNat
           _         -> return te
       }


typeExp ex gamma sigma = typeE ex gamma
    where
      typeE ex gamma  =
          case ex of 
            ETick q _ -> 
                return $ TETick q TUnit

            EUnit _ -> 
                return $ TEUnit TUnit

            ETrue _ -> 
                return $ TETrue TBool
      
            EFalse _ -> 
                return $ TEFalse TBool
      
            ENat n _ -> 
                return $ TENat n TNatConst
      
            EVar x _ -> 
                do { t <- case M.lookup x gamma of
                            Just t -> return t
                            Nothing -> fail $ "Unbound variable " ++ x 
                                              ++ " in expression " ++ (errorInfo ex)
                   ; return $ TEVar x t
                   }
      
            EApp f e _ ->
                do { (FunType tArg tRes) <- case M.lookup f sigma of
                         Just t_f -> return t_f
                         Nothing -> fail $ "Undefined function: " ++ f ++ "\n"
                                           ++ "In expression " ++ (errorInfo ex)
                   ; te <- typeWithT e gamma tArg
                   ; return $ TEApp f te tRes
                   }
      
            EUnaryOp op e _ -> 
                case op of
                  UNot -> do 
                   { te <- typeWithT e gamma TBool
                   ; return $ TEUnaryOp UNot te TBool
                   }
                  _ -> do
                   { te <- typeWithT e gamma TInt
                   ; return $ TEUnaryOp op te TInt
                   }
      
            EBinOp Eq e1 e2 _ -> 
                do { te1 <- typeE e1 gamma
                   ; let t1 = typeOf te1
                   ; if t1 `elem` [TNatConst,TNat,TInt,TBool,TUnit]
                     then do { te2 <- typeE e2 gamma
                             ; let t2 = typeOf te2
                             ; case unifyTypes t1 t2 of
                                 Nothing -> missMatch e1 t1 e2 t2
                                 Just t -> do { let t' = case t of
                                                           TNatConst -> TNat
                                                           _ -> t
                                              ; let te1' = setType te1 t'
                                              ; let te2' = setType te2 t'
                                              ; return $ TEBinOp Eq te1' te2' TBool
                                              }
                             }
                     else fail $ "Equality is only defined on base types. \n"  
                                  ++ "Inferred type: " ++ (show t1) ++ ".\n"
                                  ++ "In expression" ++ (errorInfo e1)
                   }

            EBinOp op e1 e2 _ | op `elem` [And, Or] -> 
                do { te1 <- typeWithT e1 gamma TBool
                   ; te2 <- typeWithT e2 gamma TBool
                   ; return $ TEBinOp op te1 te2 TBool
                   }

            -- integer-only operations
            EBinOp op e1 e2 _ | op `elem` [Mod, Div, Sub] ->  
                do { te1 <- typeWithT e1 gamma TInt
                   ; te2 <- typeWithT e2 gamma TInt
                   ; return $ TEBinOp op te1 te2 TInt
                   }


            -- shared int and nat operations
            -- Add, Mult, LessEq, GreaterEq, Less, and Greater.
            EBinOp op e1 e2 _ -> 
                do { te1 <- typeE e1 gamma
                   ; te2 <- typeE e2 gamma
                   ; let (t1,t2) = (typeOf te1, typeOf te2)
                   ; if t1 `elem` [TInt,TNat,TNatConst] then
                      case unifyTypes t1 t2 of
                         Nothing -> missMatch e1 t1 e2 t2
                         Just t -> do 
                            { let te1' = setType te1 t
                            ; let te2' = setType te2 t
                            ; case op of
                                _ | op `elem` [LessEq, GreaterEq, Less, Greater] ->
                                  do { te1'' <- elimNatConst $ return te1'
                                     ; te2'' <- elimNatConst $ return te2'
                                     ; return $ TEBinOp op te1'' te2'' TBool
                                     }
                                _ | op `elem` [Add, Mult] ->
                                  return $ TEBinOp op te1' te2' t
                            }
                     else fail $ "Expecting a numeric type (nat or int). \n"  
                                 ++ "Inferred type: " ++ (show t1) ++ ".\n"
                                ++ "In expression" ++ (errorInfo e1)
                   }

            ESucc e _ -> 
                do { te <- typeWithT e gamma TNat
                   ; return $ TESucc te TNat
                   }

            EToInt e _ -> 
                do { te <- typeWithT e gamma TNat
                   ; return $ TEToInt te TInt
                   }

            EMatchN e e1 x e2 _ ->
                do { te <- typeWithT e gamma TNat
                   ; te1 <- typeE e1 gamma
                   ; te2 <- typeE e2 (bind [(x,TNat)] gamma)
                   ; let (t1,t2) = (typeOf te1,typeOf te2)
                   ; case unifyTypes t1 t2 of
                       Nothing -> missMatch e1 t1 e2 t2
                       Just t -> return $ setType (TEMatchN te te1 x te2 undefined) t
                   }

            ELet x e1 e2 _ -> 
                do { te1 <- elimNatConst $ typeE e1 gamma
                   ; te2 <- elimNatConst $ typeE e2 (M.insert x (typeOf te1) gamma)
                   ; return $ TELet Normal x te1 te2 (typeOf te2)
                   }
      
            ECond e e_t e_f _ ->
                do { te <- typeWithT e gamma TBool
                   ; te_t <- elimNatConst $ typeE e_t gamma
                   ; te_f <- elimNatConst $ typeE e_f gamma
                   ; let (t_t,t_f) = (typeOf te_t, typeOf te_f)
                   ; case unifyTypes t_t t_f of
                       Nothing -> missMatch e_t t_t e_f t_f
                       Just t -> return $ setType (TECond te te_t te_f undefined) t
                   }
      
            ENil _ -> return $ TENil Normal TNil
      
            ECons e_h e_t _ ->
                do { te_h <- elimNatConst $ typeE e_h gamma
                   ; te_t <- typeList e_t gamma
                   ; let (t1,t2) = (typeOf te_h, typeOf te_t)
                   ; case unifyTypes (TList t1) t2 of
                       Nothing -> missMatch e_h (TList t1) e_t t2
                       Just t -> return $ setType (TECons Normal te_h te_t undefined) t
                   }
      
            EMatchL matchType e e1 x_h x_t e2 _ ->
                do { te <- typeList e gamma
                   ; te1 <- elimNatConst $ typeE e1 gamma
                   ; case typeOf te of
                       TNil -> fail $ "Pattern match against nil.  Cannot infer types for the variables"
                                      ++ (show x_h) ++ " and " ++ (show x_t) ++ "."
                       TList t -> 
                          do { te2 <- elimNatConst $ typeE e2 (bind [(x_h,t),(x_t,TList t)] gamma)
                             ; let (t1,t2) = (typeOf te1,typeOf te2)
                             ; case unifyTypes t1 t2 of
                                 Nothing -> missMatch e1 t1 e2 t2
                                 Just t -> return $ setType (TEMatchL matchType te te1 x_h x_t te2 undefined) t
                   }}

            ETuple es _ ->
                do { tes <- sequence [elimNatConst $ typeE e gamma | e <- es]
                   ; return $ TETuple tes (TTuple [typeOf e | e <- tes])
                   }
      
            EMatchT e xs e' _ ->
                do { te <- typeE e gamma
                   ; case typeOf te of
                       TTuple ts | length xs == length ts 
                          -> do { te' <- typeE e' (bind (zip xs ts) gamma)
                                ; return $ TEMatchT Normal te xs te' (typeOf te')
                                }
                       _  -> fail $ "Expecting " ++ (show $ length xs) ++ "-tuple " ++ ".\n"
                                    ++"Inferred type: " ++ (show (typeOf te)) ++ "\n"
                                    ++"In expression" ++ (errorInfo e)
                   }
      
            ELeaf _ -> return $ TELeaf Normal TLeaf
      
            ENodeB e e1 e2 _ ->
                do { te <- elimNatConst $ typeE e gamma
                   ; te1 <- typeBTree e1 gamma
                   ; te2 <- typeBTree e2 gamma
                   ; let (t,t1,t2) = (typeOf te, typeOf te1, typeOf te2)
                   ; case unifyTypes t1 t2 of
                       Nothing -> missMatch e1 t1 e2 t2
                       Just t' -> case unifyTypes (TBTree t) t' of
                                    Nothing  -> missMatch e (TBTree t) e1 t'
                                    Just t'' -> return $ setType (TENodeB Normal te te1 te2 undefined)  t''
                   }
      
      
            EMatchB matchType e e1 x x1 x2 e2 _ ->
                do { te <- typeBTree e gamma
                   ; case typeOf te of
                       TLeaf -> fail $ "Pattern match against leaf.  Cannot infer types for the variables"
                                        ++ (show x) ++ ", " ++ (show x1) ++ " and " ++ (show x2) ++ "."
                       TBTree t -> 
                          do { te1 <- elimNatConst $ typeE e1 gamma
                             ; te2 <- elimNatConst $ typeE e2 (bind [(x,t),(x1,TBTree t),(x2,TBTree t)] gamma)
                             ; let (t1,t2) = (typeOf te1, typeOf te2)
                             ; case unifyTypes t1 t2 of
                                 Nothing -> missMatch e1 t1 e2 t2
                                 Just t' -> return $ setType (TEMatchB matchType te te1 x x1 x2 te2 undefined) t'
                   }}


      bind bs gamma = foldl (\g (x,t) -> M.insert x t g) gamma bs

      typeWithT e gamma tExp =
          do { te <- typeE e gamma
             ; let t = typeOf te
             ; when (not $ tExp `subTypeOf` t)
                  ( fail $ "Expected type: "  ++ (show tExp) ++ "\n"
                           ++ "Inferred type: " ++ (show t) ++ "\n"
                           ++ "In expression" ++ (errorInfo e)
                  )
             ; return $ setType te tExp
             }

      typeList e gamma = 
          do { te <- typeE e gamma
             ; let t = typeOf te
             ; case t of
                 TList _ -> return te
                 TNil    -> return te
                 _       -> fail $ "Expecting list type.\n"
                                    ++"Inferred type: " ++ (show t) ++ "\n"
                                    ++"In expression" ++ (errorInfo e)
             }


      typeBTree e gamma = 
          do { te <- typeE e gamma
             ; let t = typeOf te
             ; case t of
                 TBTree _ -> return te
                 TLeaf    -> return te
                 _       -> fail $ "Expecting binary tree type.\n"
                                    ++"Inferred type: " ++ (show t) ++ "\n"
                                    ++"In expression" ++ (errorInfo e)
             }


      missMatch e1 t1 e2 t2 = 
           fail $ "Type mismatch in " ++ (sourcePos ex) ++ ".\n\n"
                  ++"Failed to match   " ++ (show t1) ++ "\n           with   " ++ (show t2) ++ "\n\n"
                  ++"In the subexpressions" ++ (errorInfo e1)
                  ++"and" ++ (errorInfo e2)
      
      errorInfo e = "\n\n" ++(show e)++ "\n\nat "++(sourcePos e)++"\n\n"





-----------------------------------
-- Check Declarations
-----------------------------------

-- analyseDecs
--   :: (Monad m) => [Dec] -> m [(FunId, [VarId], Exp, FunType)]

analyseDecs decs = 
   do { let (typDs,funDs) = partition isTypeDec decs
      ; let typDecs = sortById typDs
      ; let funDecs = sortById funDs
      ; checkDublicates typDecs
      ; checkDublicates funDecs
      ; match typDecs funDecs []
      }
   where
      isTypeDec (TypeDec _ _ _) = True
      isTypeDec _               = False
    
      sortById decs = sortBy compareIds decs
    
      compareIds d1 d2 = compare (id d2) (id d1)
    
      id dec = case dec of
                  TypeDec str _ _  -> str
                  FunDec str _ _ _ -> str
   
      checkDublicates decs  = case decs of
                                []         -> return ()
                                [d]        -> return ()
                                (d1:d2:ds) -> if (id d1) == (id d2) 
                                              then dublicateError d1 d2
                                              else checkDublicates (d2:ds)


      match []     []     acc = return acc
      match (t:_)  []     _   = missingDefError (id t)
      match []     (d:_)  _   = missingTypeError (id d)
      match (t:ts) (d:ds) acc = case compareIds t d of
                                   EQ -> let (TypeDec id a _) = t
                                             (FunType args _)  = a
                                             (FunDec _ xs e _) = d
                                         in if (length xs) > 1 && (numArgs args) /= (length xs)
                                               then wrongArgNumError id a
                                               else match ts ds ( (id,xs,e,a) :acc)
                                   LT -> missingDefError (id t)
                                   GT -> missingTypeError (id d)
      
      numArgs (TTuple ts) = length ts
      numArgs _           = 1

      dublicateError d1 d2 = let str = case d1 of
                                          TypeDec f _ _  -> "Type of function " ++ f ++ " declared twice "
                                          FunDec f _ _ _ -> "Function " ++ f ++  " defined twice "
                             in
                             fail $ str ++ "at " ++( sourcePos d1) ++ " and " ++ (sourcePos d2) ++ "."

      missingTypeError name = fail $ "Missing type for function " ++ name ++ "."
      missingDefError name = fail $ "Missing definition for function " ++ name ++ "."

      wrongArgNumError name expectedT = 
           fail $ "Wrong number of arguments in function " ++ name 
                   ++ ".  Expecting type: " ++ name ++ ": " ++ (show expectedT) ++ "."
